"""Module to manange AiiDAlab configuration."""
from os import getenv
from pathlib import Path

import click
import toml

CONFIG_PATH = Path.home() / "aiidalab.toml"
_CONFIG = toml.loads(CONFIG_PATH.read_text()) if CONFIG_PATH.is_file() else dict()
_DEVELOP_MODE = _CONFIG.get("develop", False)

if _DEVELOP_MODE:  # Warn developer that the mode is enabled.
    lines = [f"Executing '{__name__.split('.')[0]}' in DEVELOP mode."]
    config_ = {k: v for k, v in _CONFIG.items() if k != "develop"}
    if config_:  # The config does not only contain the 'develop' key.
        lines.append(f"{CONFIG_PATH!s} (takes precedence):")
        lines.extend(toml.dumps(config_).splitlines())
    click.secho("\n".join([f"\U0001F6A7  {line}" for line in lines]), fg="yellow")


def _as_env_var_name(key):
    return "AIIDALAB_" + key.upper()


def _get_config_value(key, default=None):
    """Return config value from configuration source.

    The standard configuration source order is:

        1. environment variables
        2. local configuration

    In 'develop' mode the order is reversed to simplify local
    override of configuration values.

    The convention for environment variables it to prefix keys
    with `AIIDALAB_` and to convert all letters to uppercase.
    For example, the configuration key `registry` is interpreted as
    `AIIDALAB_REGISTRY` when sourced as an environment variable.
    """
    if _DEVELOP_MODE:
        try:
            return _CONFIG[key]
        except KeyError:
            return getenv(_as_env_var_name(key), default)
    return getenv(_as_env_var_name(key), _CONFIG.get(key, default))


AIIDALAB_HOME = _get_config_value("home", "/project")
AIIDALAB_APPS = _get_config_value("apps", "/project/apps")
AIIDALAB_SCRIPTS = _get_config_value("scripts", "/opt")
AIIDALAB_REGISTRY = _get_config_value(
    "registry", "https://aiidalab.github.io/aiidalab-registry/api/v1"
)
AIIDALAB_DEFAULT_GIT_BRANCH = _get_config_value("default_git_branch", "master")



"""Helpful utilities for the AiiDAlab tools."""

import json
import logging
import subprocess
import sys
import time
from collections import defaultdict
from dataclasses import asdict
from functools import wraps
from pathlib import Path
from subprocess import run
from threading import Lock
from urllib.parse import urlsplit, urlunsplit

import requests
from cachetools import TTLCache, cached
from packaging.utils import canonicalize_name

from .config import AIIDALAB_REGISTRY
from .environment import Environment
from .fetch import fetch_from_url
from .metadata import Metadata

logger = logging.getLogger(__name__)
FIND_INSTALLED_PACKAGES_CACHE = TTLCache(maxsize=32, ttl=60)

# Warning: try-except is a fix for Quantum Mobile release v19.03.0 where
# requests_cache is not installed.
try:
    import requests_cache
except ImportError:
    logger.warning(
        "The requests_cache package is missing.  Requests made to the app "
        "registry will not be cached!"
    )
else:
    # Install global cache for all requests.

    # The cache file is placed within the user home directory.
    cache_file = Path.home().joinpath(".cache", "requests", "cache")
    cache_file.parent.mkdir(parents=True, exist_ok=True)

    # The cache is configured to avoid spamming the index server with requests
    # that are made in rapid succession and also serves as a fallback in case
    # that the index server is not reachable.
    requests_cache.install_cache(
        cache_name=str(cache_file),
        backend="sqlite",
        expire_after=60,  # seconds
        old_data_on_error=True,
    )


def load_app_registry_index():
    """Load apps' information from the AiiDAlab registry."""
    try:
        return requests.get(f"{AIIDALAB_REGISTRY}/apps_index.json").json()
    except (ValueError, requests.ConnectionError) as error:
        raise RuntimeError(f"Unable to load registry index: '{error}'")


def load_app_registry_entry(app_id):
    """Load registry enty for app with app_id."""
    try:
        return requests.get(f"{AIIDALAB_REGISTRY}/apps/{app_id}.json").json()
    except (ValueError, requests.ConnectionError):
        logger.debug(f"Unable to load registry entry for app with id '{app_id}'.")
        return None


class PEP508CompliantUrl(str):
    """Represents a PEP 508 compliant URL."""

    pass


def parse_app_repo(url, metadata_fallback=None):
    """Parse an app repo for metadata and other information.

    Use this function to parse a local or remote app repository for the app
    metadata and environment specification.

    Examples:

    For a local app repository, provide the absolute or relative path:

        url="/path/to/aiidalab-hello-world"

    For a remote app repository, provide a PEP 508 compliant URL, for example:

        url="git+https://github.com/aiidalab/aiidalab-hello-world.git@v1.0.0"
    """
    with fetch_from_url(url) as repo:
        try:
            metadata = asdict(Metadata.parse(repo))
        except TypeError as error:
            logger.debug(f"Failed to parse metadata for '{url}': {error}")
            metadata = metadata_fallback

        return {
            "metadata": metadata,
            "environment": asdict(Environment.scan(repo)),
        }


class throttled:  # pylint: disable=invalid-name
    """Decorator to throttle calls to a function to a specified rate.

    The throttle is specific to the first argument of the wrapped
    function. That means for class methods it is specific to each
    instance.

    Adapted from: https://gist.github.com/gregburek/1441055

    """

    def __init__(self, calls_per_second=1):
        self.calls_per_second = calls_per_second
        self.last_start = defaultdict(lambda: -1)
        self.locks = defaultdict(Lock)

    def __call__(self, func):
        """Return decorator function."""

        @wraps(func)
        def wrapped(instance, *args, **kwargs):
            if self.last_start[hash(instance)] >= 0:
                elapsed = time.perf_counter() - self.last_start[hash(instance)]
                to_wait = 1.0 / self.calls_per_second - elapsed
                if to_wait > 0:
                    locked = self.locks[hash(instance)].acquire(blocking=False)
                    if locked:
                        try:
                            time.sleep(to_wait)
                        finally:
                            self.locks[hash(instance)].release()
                    else:
                        return None  # drop

            self.last_start[hash(instance)] = time.perf_counter()
            return func(instance, *args, **kwargs)

        return wrapped


class Package:
    """Helper class to check whether a given package fulfills a requirement."""

    def __init__(self, name, version):
        self.name = name
        self.version = version

    def __str__(self):
        return f"{type(self).__name__}({self.name}, {self.version})"

    def fulfills(self, requirement):
        """Returns True if this entry fullfills the requirement."""
        return (
            canonicalize_name(self.name) == canonicalize_name(requirement.name)
            and self.version in requirement.specifier
        )


@cached(cache=FIND_INSTALLED_PACKAGES_CACHE)
def find_installed_packages(python_bin=None):
    """Return all currently installed packages."""
    if python_bin is None:
        python_bin = sys.executable
    output = run(
        [python_bin, "-m", "pip", "list", "--format=json"],
        encoding="utf-8",
        capture_output=True,
    ).stdout
    return [
        Package(name=package["name"], version=package["version"])
        for package in json.loads(output)
    ]


def split_git_url(git_url):
    """Split the base url and the ref pointer of a git url.

    For example: git+https://example.com/app.git@v1 is split into and returned
    as tuple: (git+https://example.com/app.git, v1)
    """

    parsed_url = urlsplit(git_url)
    if "@" in parsed_url.path:
        path, ref = parsed_url.path.rsplit("@", 1)
    else:
        path, ref = parsed_url.path, None
    return urlunsplit(parsed_url._replace(path=path)), ref


def this_or_only_subdir(path):
    members = list(path.iterdir())
    return members[0] if len(members) == 1 and members[0].is_dir() else path


def run_pip_install(*args, python_bin=sys.executable):
    return subprocess.Popen(
        [python_bin, "-m", "pip", "install", *args],
        bufsize=1,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )


def run_reentry_scan():
    return subprocess.Popen(
        ["reentry", "scan"],
        bufsize=1,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )


def run_verdi_daemon_restart():
    return subprocess.Popen(
        ["verdi", "daemon", "restart"],
        bufsize=1,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )



# -*- coding: utf-8 -*-
"""Core data classes for the app registry."""
import json
from dataclasses import dataclass, fields

import jsonschema
import pkg_resources

from .util import load_json


@dataclass
class AppRegistrySchemas:
    """The app registry JSON-schema objects."""

    app: dict
    apps: dict
    apps_index: dict
    categories: dict
    environment: dict
    metadata: dict

    @classmethod
    def from_path(cls, path):
        return cls(
            **{
                field.name: load_json(path.joinpath(f"{field.name}.schema.json"))
                for field in fields(cls)
            }
        )

    @classmethod
    def from_package(cls):
        return cls(
            **{
                field.name: json.loads(
                    pkg_resources.resource_string(
                        "aiidalab.registry", f"schemas/{field.name}.schema.json"
                    )
                )
                for field in fields(cls)
            }
        )


@dataclass
class AppRegistryData:
    """The app registry data objects (apps and categories)."""

    apps: dict
    categories: dict

    def validate(self, schemas: AppRegistrySchemas):
        """Validate the registry data against the provided registry schemas."""
        jsonschema.validate(instance=self.apps, schema=schemas.apps)
        jsonschema.validate(instance=self.categories, schema=schemas.categories)



# -*- coding: utf-8 -*-
"""Module to manage AiiDAlab apps."""

import errno
import io
import logging
import os
import shutil
import sys
import tarfile
import tempfile
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from enum import Enum, Flag, auto
from itertools import repeat
from pathlib import Path
from subprocess import CalledProcessError
from threading import Thread
from time import sleep
from urllib.parse import urldefrag, urlsplit, urlunsplit
from uuid import uuid4

import requests
import traitlets
from dulwich.errors import NotGitRepository
from packaging.requirements import Requirement
from packaging.version import parse
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver

from .config import AIIDALAB_APPS
from .environment import Environment
from .git_util import GitManagedAppRepo as Repo
from .git_util import git_clone
from .metadata import Metadata
from .utils import (
    FIND_INSTALLED_PACKAGES_CACHE,
    PEP508CompliantUrl,
    find_installed_packages,
    load_app_registry_entry,
    load_app_registry_index,
    run_pip_install,
    run_reentry_scan,
    run_verdi_daemon_restart,
    split_git_url,
    this_or_only_subdir,
    throttled,
)

logger = logging.getLogger(__name__)


# A version is usually of type str, but it can also be a value
# of this Enum to indicate special app states in which the
# version cannot be determined, e.g., because the app is in a
# detached state, or because the app is not installed at all.
class AppVersion(Enum):
    UNKNOWN = auto()
    NOT_INSTALLED = auto()


class AppRemoteUpdateStatus(Flag):
    NOT_REGISTERED = auto()
    UP_TO_DATE = auto()
    UPDATE_AVAILABLE = auto()
    CANNOT_REACH_REGISTRY = auto()
    DETACHED = auto()


@dataclass
class _AiidaLabApp:

    metadata: dict
    name: str
    path: Path
    releases: dict = field(default_factory=dict)

    @classmethod
    def from_registry_entry(cls, path, registry_entry):
        return cls(
            path=path,
            **{
                key: value
                for key, value in registry_entry.items()
                if key in ("metadata", "name", "releases")
            },
        )

    @classmethod
    def _registry_entry_from_path(cls, path):
        try:
            return {
                "name": path.stem,
                "metadata": asdict(Metadata.parse(path)),
                "releases": None,
            }
        except (TypeError, ValueError):
            logger.debug(f"Unable to parse metadata from '{path}'")
            return {
                "name": path.stem,
                "metadata": dict(title=path.stem, description=""),
                "releases": None,
            }

    @classmethod
    def from_id(cls, app_id, registry_entry=None, apps_path=None):
        if apps_path is None:
            apps_path = AIIDALAB_APPS

        app_path = Path(apps_path).joinpath(app_id)

        if registry_entry is None:
            local_registry_entry = cls._registry_entry_from_path(app_path)
            remote_registry_entry = load_app_registry_entry(app_id)
            registry_entry = remote_registry_entry or local_registry_entry

        return cls.from_registry_entry(path=app_path, registry_entry=registry_entry)

    @property
    def is_registered(self):
        try:
            app_registry_index = load_app_registry_index()
        except RuntimeError as error:
            logger.warning(str(error))
            return None
        else:
            return self.name in app_registry_index["apps"]

    @property
    def _repo(self):
        try:
            return Repo(str(self.path))
        except NotGitRepository:
            return None

    def installed_version(self):
        if self._repo and self.is_registered:
            if self.dirty():
                return AppVersion.UNKNOWN
            else:
                try:
                    head_commit = self._repo.head().decode()
                    versions_by_commit = {
                        split_git_url(release["url"])[1]: version
                        for version, release in self.releases.items()
                        if urlsplit(release["url"]).scheme.startswith("git+")
                    }
                    return versions_by_commit.get(
                        head_commit, self.metadata.get("version", AppVersion.UNKNOWN)
                    )
                except Exception as error:
                    logger.debug(
                        f"Encountered error while determining version: {error}"
                    )
                    return AppVersion.UNKNOWN
        elif self.path.exists():
            return self.metadata.get("version", AppVersion.UNKNOWN)
        return AppVersion.NOT_INSTALLED

    def available_versions(self):
        if self.is_registered:
            yield from sorted(self.releases, key=parse, reverse=True)

    def dirty(self):
        if self._repo:
            return self._repo.dirty()

    def is_installed(self):
        """The app is installed if the corresponding folder is present."""
        return self.path.exists()

    def remote_update_status(self, prereleases=False):
        """Determine the remote update satus.

        Arguments:
            prereleases (Bool):
                Set to True to include available preleases. Defaults to False.
        Returns:
            AppRemoteUpdateStatus
        """
        if self.is_installed():

            # Check whether app is registered.
            if self.is_registered is None:
                return AppRemoteUpdateStatus.CANNOT_REACH_REGISTRY

            if self.is_registered is False:
                return AppRemoteUpdateStatus.NOT_REGISTERED

            # Check whether the locally installed version is a registered release.
            installed_version = self.installed_version()
            if installed_version is AppVersion.UNKNOWN:
                return AppRemoteUpdateStatus.DETACHED

            # Check whether the locally installed version is the latest release.
            available_versions = [
                version
                for version in sorted(self.releases, key=parse, reverse=True)
                if prereleases or not parse(version).is_prerelease
            ]
            if len(available_versions) and installed_version != available_versions[0]:
                return AppRemoteUpdateStatus.UPDATE_AVAILABLE

            # App must be up-to-date.
            return AppRemoteUpdateStatus.UP_TO_DATE

        return AppRemoteUpdateStatus(0)  # app is not installed

    def _move_to_trash(self):
        trash_path = Path.home().joinpath(".trash", f"{self.name}-{uuid4()!s}")
        if self.path.exists():
            trash_path.parent.mkdir(parents=True, exist_ok=True)
            self.path.rename(trash_path)
            return trash_path

    def _restore_from(self, trash_path):
        self._move_to_trash()
        trash_path.rename(self.path)

    def uninstall(self):
        self._move_to_trash()

    def find_matching_releases(self, specifier, prereleases=None):
        matching_releases = list(
            specifier.filter(self.releases, prereleases=prereleases)
        )
        # Sort by intrinsic order (e.g. 1.1.0 -> 1.0.1 -> 1.0.0 and so on)
        matching_releases.sort(key=parse, reverse=True)
        return matching_releases

    @staticmethod
    def _find_incompatibilities_python(requirements, python_bin):
        packages = find_installed_packages(python_bin)
        for requirement in map(Requirement, requirements):
            f = [p for p in packages if p.fulfills(requirement)]
            if not any(f):
                yield requirement

    def find_incompatibilities(self, version, python_bin=None):
        if python_bin is None:
            python_bin = sys.executable

        if not self.is_registered:
            assert version == self.installed_version()
            environment = asdict(Environment.scan(self.path))
        else:
            environment = self.releases[version].get("environment")

        for key, spec in environment.items():
            if key == "python_requirements":
                yield from zip(
                    repeat("python"),
                    self._find_incompatibilities_python(spec, python_bin),
                )
            else:
                raise ValueError(f"Unknown eco-system '{key}'")

    def is_compatible(self, version, python_bin=None):
        return not any(self.find_incompatibilities(version, python_bin))

    def _install_dependencies(self, python_bin, stdout=None):
        """Try to install the app dependencies with pip (if specified)."""

        def _pip_install(*args, stdout=None):
            # The implementation of this function is taken and adapted from:
            # https://www.endpoint.com/blog/2015/01/getting-realtime-output-using-python/
            stdout = stdout or sys.stdout

            # Install package dependencies.
            process = run_pip_install(*args, python_bin=python_bin)
            for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
                stdout.write(line)

            # AiiDA plugins require reentry run to be found by AiiDA.
            process = run_reentry_scan()
            for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
                stdout.write(line)

            # Restarting the AiiDA daemon to import newly installed plugins.
            process = run_verdi_daemon_restart()
            for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
                stdout.write(line)

        for path in (self.path.joinpath(".aiidalab"), self.path):
            if path.exists():
                try:
                    if path.joinpath("setup.py").is_file():
                        _pip_install(
                            "--use-feature=in-tree-build", str(path), stdout=stdout
                        )
                    elif path.joinpath("requirements.txt").is_file():
                        _pip_install(
                            f"--requirement={path.joinpath('requirements.txt')}",
                            stdout=stdout,
                        )
                    else:
                        logger.warning(
                            f"Warning: App '{self.name}' does not declare any dependencies."
                        )
                    break
                except CalledProcessError:
                    raise RuntimeError("Failed to install dependencies.")

    def _install_from_path(self, path):
        if path.is_dir():
            shutil.copytree(this_or_only_subdir(path), self.path)
        else:
            with tempfile.TemporaryDirectory() as tmp_dir:
                with tarfile.open(path) as tar_file:
                    tar_file.extractall(path=tmp_dir)
                    self._install_from_path(Path(tmp_dir))

    def _install_from_https(self, url):
        response = requests.get(url, stream=True)
        response.raise_for_status()
        content = response.content

        with tempfile.NamedTemporaryFile() as tmp_file:
            tmp_file.write(content)
            tmp_file.flush()
            self._install_from_path(Path(tmp_file.name))

    def _install_from_git_repository(self, git_url):
        if urldefrag(git_url).fragment:
            raise NotImplementedError(
                "Path specification via fragment not yet supported."
            )
        base_url, ref = split_git_url(git_url)
        git_clone(base_url, ref, self.path)

    def install(
        self,
        version=None,
        python_bin=None,
        install_dependencies=True,
        stdout=None,
        prereleases=False,
    ):
        if version is None:
            try:
                version = [
                    version
                    for version in sorted(self.releases, key=parse)
                    if prereleases or not parse(version).is_prerelease
                ][-1]
            except IndexError:
                raise ValueError(f"No versions available for '{self}'.")
        if python_bin is None:
            python_bin = sys.executable

        if isinstance(version, PEP508CompliantUrl):
            url = version
        else:
            url = self.releases[version]["url"]

        trash_path = self._move_to_trash()

        split_url = urlsplit(url)
        try:
            if split_url.scheme in ("", "file"):
                self._install_from_path(Path(split_url.path))
            elif split_url.scheme == "https":
                self._install_from_https(url)
            elif split_url.scheme == "git+file":
                self._install_from_git_repository(
                    urlunsplit(split_url._replace(scheme="file"))
                )
            elif split_url.scheme == "git+https":
                self._install_from_git_repository(
                    urlunsplit(split_url._replace(scheme="https"))
                )
            else:
                raise NotImplementedError(
                    "Unsupported scheme: {split_url.scheme} ({url})"
                )

            # Install dependencies
            if install_dependencies:
                self._install_dependencies(python_bin, stdout=stdout)
        except RuntimeError as error:
            try:
                if trash_path is None:
                    # App, was not previously installed, just remove it.
                    logger.info("Removing partially installed app.")
                    self._move_to_trash()
                else:
                    # Attempt rollback to previous version.
                    logger.info("Performing rollback to previously installed version.")
                    self._restore_from(trash_path)
            except RuntimeError as error:
                logger.warning(f"Rollback failed due to error: {error}!")
            finally:
                raise RuntimeError(
                    f"Failed to install '{self.name}' (version={version}) at '{self.path}'"
                    f", due to error: {error}"
                )


class AppNotInstalledException(Exception):
    pass


class AiidaLabAppWatch:
    """Watch to monitor the app installation status.

    Create a watch instance to monitor the installation status of an
    AiiDAlab app. This is achieved by monitoring the app repository
    for existance and changes.

    Arguments:
        app (AiidaLabApp):
            The AiidaLab app to monitor.
    """

    class AppPathFileSystemEventHandler(FileSystemEventHandler):
        """Internal event handeler for app path file system events."""

        def __init__(self, app):
            self.app = app

        def on_any_event(self, event):
            """Refresh app for any event."""
            self.app.refresh_async()

    def __init__(self, app):
        self.app = app

        self._started = False
        self._monitor_thread = None
        self._observer = None

    def __repr__(self):
        return f"<{type(self).__name__}(app={self.app!r})>"

    def _start_observer(self):
        """Start the directory observer thread.

        The ._observer thread is controlled by the ._monitor_thread.
        """
        assert os.path.isdir(self.app.path)
        assert self._observer is None or not self._observer.isAlive()

        event_handler = self.AppPathFileSystemEventHandler(self.app)

        self._observer = Observer()
        self._observer.schedule(event_handler, self.app.path, recursive=True)
        try:
            self._observer.start()
        except OSError as error:
            if error.errno in (errno.ENOSPC, errno.EMFILE) and "inotify" in str(error):
                # We reached the inotify watch limit, using polling-based fallback observer.
                self._observer = PollingObserver()
                self._observer.schedule(event_handler, self.app.path, recursive=True)
                self._observer.start()
            else:  # reraise unrelated error
                raise error

    def _stop_observer(self):
        """Stop the directory observer thread.

        The ._observer thread is controlled by the ._monitor_thread.
        """
        assert self._observer is not None
        self._observer.stop()

    def start(self):
        """Watch the app repository for file system events.

        The app state is refreshed automatically for all events.
        """
        if self._started:
            raise RuntimeError(
                f"Instances of {type(self).__name__} can only be started once."
            )

        if self._monitor_thread is None:

            def check_path_exists_changed():
                is_dir = os.path.isdir(self.app.path)
                while not self._monitor_thread.stop_flag:
                    switched = is_dir != os.path.isdir(self.app.path)
                    if switched:
                        is_dir = not is_dir
                        self.app.refresh()

                    if is_dir:
                        if self._observer is None or not self._observer.isAlive():
                            self._start_observer()
                    elif self._observer and self._observer.isAlive():
                        self._stop_observer()

                    sleep(1)

                # stop-flag set, stopping observer...
                if self._observer:
                    self._observer.stop()

            self._monitor_thread = Thread(target=check_path_exists_changed)
            self._monitor_thread.stop_flag = False
            self._monitor_thread.start()

        self._started = True

    def stop(self):
        """Stop watching the app repository for file system events."""
        if self._monitor_thread is not None:
            self._monitor_thread.stop_flag = True

    def is_alive(self):
        """Return True if this watch is still alive."""
        return self._monitor_thread and self._monitor_thread.is_alive()

    def join(self, timeout=None):
        """Join the watch after stopping.

        This function will timeout if a timeout argument is provided. Use the
        is_alive() function to determien whether the watch was stopped within
        the given timout.
        """
        if self._monitor_thread is not None:
            self._monitor_thread.join(timeout=timeout)


class AiidaLabApp(traitlets.HasTraits):
    """Manage installation status of an AiiDAlab app.

    Arguments:

        name (str):
            Name of the Aiida lab app.
        app_data (dict):
            Dictionary containing the app metadata.
        aiidalab_apps_path (str):
            Path to directory at which the app is expected to be installed.
        watch (bool):
            If true (default), automatically watch the repository for changes.
    """

    path = traitlets.Unicode(allow_none=True, readonly=True)
    install_info = traitlets.Unicode()

    available_versions = traitlets.List(traitlets.Unicode)
    installed_version = traitlets.Union(
        [traitlets.Unicode(), traitlets.UseEnum(AppVersion)]
    )
    remote_update_status = traitlets.UseEnum(
        AppRemoteUpdateStatus, readonly=True, allow_none=True
    )
    has_prereleases = traitlets.Bool()
    include_prereleases = traitlets.Bool()

    busy = traitlets.Bool(readonly=True)
    detached = traitlets.Bool(readonly=True, allow_none=True)
    compatible = traitlets.Bool(readonly=True, allow_none=True)
    compatibility_info = traitlets.Dict()

    def __init__(self, name, app_data, aiidalab_apps_path, watch=True):
        self._app = _AiidaLabApp.from_id(
            name, registry_entry=app_data, apps_path=aiidalab_apps_path
        )
        super().__init__()

        self.name = self._app.name
        self.logo = self._app.metadata["logo"]
        self.categories = self._app.metadata["categories"]
        self.is_installed = self._app.is_installed
        self.path = str(self._app.path)
        self.refresh_async()

        if watch:
            self._watch = AiidaLabAppWatch(self)
            self._watch.start()
        else:
            self._watch = None

    def __str__(self):
        return f"<AiidaLabApp name='{self._app.name}'>"

    @traitlets.default("include_prereleases")
    def _default_include_prereleases(self):
        "Provide default value for include_prereleases trait." ""
        return False

    @traitlets.observe("include_prereleases")
    def _observe_include_prereleases(self, change):
        if change["old"] != change["new"]:
            self.refresh()

    @traitlets.default("detached")
    def _default_detached(self):
        """Provide default value for detached traitlet."""
        if self.is_installed():
            return self._app.dirty() or self._installed_version() is AppVersion.UNKNOWN
        return None

    @traitlets.default("busy")
    def _default_busy(self):  # pylint: disable=no-self-use
        return False

    @contextmanager
    def _show_busy(self):
        """Apply this decorator to indicate that the app is busy during execution."""
        self.set_trait("busy", True)
        try:
            yield
        finally:
            self.set_trait("busy", False)

    def in_category(self, category):
        # One should test what happens if the category won't be defined.
        return category in self.categories

    def _has_git_repo(self):
        """Check if the app has a .git folder in it."""
        try:
            Repo(self.path)
            return True
        except NotGitRepository:
            return False

    def install_app(self, version=None, stdout=None):
        """Installing the app."""
        with self._show_busy():
            self._app.install(
                version=version, stdout=stdout, prereleases=self.include_prereleases
            )
            FIND_INSTALLED_PACKAGES_CACHE.clear()
            self.refresh()
            return self._installed_version()

    def update_app(self, _=None, stdout=None):
        """Perform app update."""
        with self._show_busy():
            # Installing with version=None automatically selects latest
            # available version.
            version = self.install_app(version=None, stdout=stdout)
            FIND_INSTALLED_PACKAGES_CACHE.clear()
            self.refresh()
            return version

    def uninstall_app(self, _=None):
        """Perfrom app uninstall."""
        # Perform uninstall process.
        with self._show_busy():
            self._app.uninstall()
            self.refresh()

    def _installed_version(self):
        """Determine the currently installed version."""
        return self._app.installed_version()

    @traitlets.default("compatible")
    def _default_compatible(self):  # pylint: disable=no-self-use
        return None

    def _is_compatible(self, app_version):
        """Determine whether the specified version is compatible."""
        try:
            incompatibilities = dict(
                self._app.find_incompatibilities(version=app_version)
            )
            self.compatibility_info = {
                app_version: [
                    f"({eco_system}) {requirement}"
                    for eco_system, requirement in incompatibilities.items()
                ]
            }

            return not any(incompatibilities)
        except KeyError:
            return None  # compatibility indetermined for given version

    def _remote_update_status(self):
        """Determine whether there are updates available.

        For this the app must be installed in a known version and there must be
        available (and compatible) versions.
        """
        installed_version = self._installed_version()
        if installed_version not in (AppVersion.UNKNOWN, AppVersion.NOT_INSTALLED):
            available_versions = list(self.available_versions)
            if len(available_versions):
                return self._installed_version() != available_versions[0]
        return False

    def _refresh_versions(self):
        self.installed_version = self._app.installed_version()
        self.include_prereleases = self.include_prereleases or (
            isinstance(self.installed_version, str)
            and parse(self.installed_version).is_prerelease
        )

        all_available_versions = list(self._app.available_versions())
        self.has_prereleases = any(
            parse(version).is_prerelease for version in all_available_versions
        )
        if self._app.is_registered:
            self.available_versions = [
                version
                for version in all_available_versions
                if self.include_prereleases or not parse(version).is_prerelease
            ]

    @throttled(calls_per_second=1)
    def refresh(self):
        """Refresh app state."""
        with self._show_busy():
            with self.hold_trait_notifications():
                self._refresh_versions()
                self.set_trait(
                    "compatible", self._is_compatible(self.installed_version)
                )
                self.set_trait(
                    "remote_update_status",
                    self._app.remote_update_status(
                        prereleases=self.include_prereleases
                    ),
                )
                self.set_trait(
                    "detached",
                    (self.installed_version is AppVersion.UNKNOWN)
                    if (self._has_git_repo() and self._app.is_registered)
                    else None,
                )

    def refresh_async(self):
        """Asynchronized (non-blocking) refresh of the app state."""
        refresh_thread = Thread(target=self.refresh)
        refresh_thread.start()

    @property
    def metadata(self):
        """Return metadata dictionary. Give the priority to the local copy (better for the developers)."""
        return self._app.metadata

    def _get_from_metadata(self, what):
        """Get information from metadata."""
        try:
            return "{}".format(self._app.metadata[what])
        except KeyError:
            return f'Field "{what}" is not present in app metadata.'

    @property
    def authors(self):
        return self._get_from_metadata("authors")

    @property
    def description(self):
        return self._get_from_metadata("description")

    @property
    def title(self):
        return self._get_from_metadata("title")

    @property
    def url(self):
        """Provide explicit link to Git repository."""
        return self._get_from_metadata("external_url")

    @property
    def more(self):
        return """<a href=./single_app.ipynb?app={}>Manage App</a>""".format(self.name)

    @property
    def _repo(self):
        """Returns Git repository."""
        if not self.is_installed():
            raise AppNotInstalledException("The app is not installed")
        return Repo(self.path)



# -*- coding: utf-8 -*-
"""Utility module for git-managed AiiDAlab apps."""
import locale
import os
import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from subprocess import CalledProcessError, run
from urllib.parse import urldefrag

from dulwich.porcelain import branch_list, status
from dulwich.repo import Repo


class BranchTrackingStatus(Enum):
    """Descripe the tracking status of a branch."""

    BEHIND = -1
    EQUAL = 0
    AHEAD = 1
    DIVERGED = 2


class GitManagedAppRepo(Repo):
    """Utility class to simplify management of git-based apps."""

    def list_branches(self):
        """List all repository branches."""
        return branch_list(self)

    def branch(self):
        """Return the current branch.

        Raises RuntimeError if the repository is in a detached HEAD state.
        """
        branches = self._get_branch_for_ref(b"HEAD")
        if branches:
            return branches[0]
        raise RuntimeError("In detached HEAD state.")

    def get_tracked_branch(self, branch=None):
        """Return the tracked branch for a given branch or None if the branch is not tracking."""
        if branch is None:
            branch = self.branch()

        cfg = self.get_config()
        try:
            remote = cfg[(b"branch", branch)][b"remote"]
            merge = cfg[(b"branch", branch)][b"merge"]
            pattern = rb"refs\/heads"
            remote_ref = b"refs/remotes/" + remote + re.sub(pattern, b"", merge)
            return remote_ref
        except KeyError:
            return None

    def dirty(self):
        """Check if there are likely local user modifications to the app repository."""
        status_ = status(self)
        return bool(any(bool(_) for _ in status_.staged.values()) or status_.unstaged)

    def update_available(self):
        """Check whether there non-pulled commits on the tracked branch."""
        return (
            self.get_branch_tracking_status(self.branch())
            is BranchTrackingStatus.BEHIND
        )

    def get_branch_tracking_status(self, branch):
        """Return the tracking status of branch."""
        tracked_branch = self.get_tracked_branch(branch)
        if tracked_branch:
            ref = b"refs/heads/" + branch

            # Check if local branch points to same commit as tracked branch:
            if self.refs[ref] == self.refs[tracked_branch]:
                return BranchTrackingStatus.EQUAL

            # Check if local branch is behind the tracked branch:
            for commit in self.get_walker(self.refs[tracked_branch]):
                if commit.commit.id == self.refs[ref]:
                    return BranchTrackingStatus.BEHIND

            # Check if local branch is ahead of tracked branch:
            for commit in self.get_walker(self.refs[ref]):
                if commit.commit.id == self.refs[tracked_branch]:
                    return BranchTrackingStatus.AHEAD

            return BranchTrackingStatus.DIVERGED

        return None

    def _get_branch_for_ref(self, ref):
        """Get the branch name for a given reference."""
        pattern = rb"refs\/heads\/"
        return [
            re.sub(pattern, b"", ref)
            for ref in self.refs.follow(ref)[0]
            if re.match(pattern, ref)
        ]


def git_clone(url, commit, path):
    try:
        run(
            ["git", "clone", str(url), str(path)],
            capture_output=True,
            encoding="utf-8",
            check=True,
        )
        if commit is not None:
            run(
                ["git", "checkout", str(commit)],
                capture_output=True,
                encoding="utf-8",
                check=True,
                cwd=str(path),
            )
    except CalledProcessError as error:
        raise RuntimeError(error.stderr)


@dataclass
class GitPath(os.PathLike):
    """Utility class to operate on git objects like path objects."""

    repo: Path
    commit: str
    path: Path = Path(".")

    def __fspath__(self):
        return str(Path(self.repo).joinpath(self.path))

    def joinpath(self, *other):
        return type(self)(
            repo=self.repo, path=self.path.joinpath(*other), commit=self.commit
        )

    def resolve(self, strict=False):
        return type(self)(
            repo=self.repo,
            path=Path(self.repo)
            .joinpath(self.path)
            .resolve(strict=strict)
            .relative_to(Path(self.repo).resolve()),
            commit=self.commit,
        )

    def _get_type(self):
        # Git expects that a current directory path ("." or "./") is
        # represented with a trailing slash for this function.
        path = "./" if self.path == Path() else self.path

        try:
            return (
                run(
                    ["git", "cat-file", "-t", f"{self.commit}:{path}"],
                    cwd=self.repo,
                    check=True,
                    capture_output=True,
                )
                .stdout.decode(errors="ignore")
                .strip()
            )
        except CalledProcessError as error:
            if error.returncode == 128:
                return None
            raise

    def is_file(self):
        return self._get_type() == "blob"

    def is_dir(self):
        return self._get_type() == "tree"

    def read_bytes(self):
        try:
            return run(
                ["git", "show", f"{self.commit}:{self.path}"],
                cwd=os.fspath(self.repo),
                check=True,
                capture_output=True,
            ).stdout
        except CalledProcessError as error:
            error_message = error.stderr.decode(errors="ignore").strip()
            if re.match(
                f"fatal: Path '{re.escape(str(self.path))}' (exists on disk, but not in|does not exist)",
                error_message,
                flags=re.IGNORECASE,
            ):
                raise FileNotFoundError(f"{self.commit}:{self.path}")
            elif re.match(
                "fatal: Invalid object name", error_message, flags=re.IGNORECASE
            ):
                raise ValueError(f"Unknown commit: {self.commit}")
            else:
                raise  # unexpected error

    def read_text(self, encoding=None, errors=None):
        if encoding is None:
            encoding = locale.getpreferredencoding(False)
        if errors is None:
            errors = "strict"
        return self.read_bytes().decode(encoding=encoding, errors=errors)


class GitRepo(Repo):
    def get_current_branch(self):
        try:
            branch = run(
                ["git", "branch", "--show-current"],
                cwd=Path(self.path),
                check=True,
                capture_output=True,
                encoding="utf-8",
            ).stdout
        except CalledProcessError as error:
            if error.returncode == 129:
                raise RuntimeError("This function equires at least git version 2.22.")
            raise
        if not branch:
            raise RuntimeError(
                "Unable to determine current branch name, likely in detached HEAD state."
            )
        return branch.strip()

    def get_commit_for_tag(self, tag):
        return self.get_peeled(f"refs/tags/{tag}".encode()).decode()

    def get_merged_tags(self, branch):
        for branch_ref in [f"refs/heads/{branch}", f"refs/remotes/origin/{branch}"]:
            if branch_ref.encode() in self.refs:
                yield from run(
                    ["git", "tag", "--merged", branch_ref],
                    cwd=self.path,
                    check=True,
                    capture_output=True,
                    encoding="utf-8",
                ).stdout.splitlines()
                break
        else:
            raise ValueError(f"Not a branch: {branch}")

    def rev_list(self, rev_selection):
        return run(
            ["git", "rev-list", rev_selection],
            cwd=self.path,
            check=True,
            encoding="utf-8",
            capture_output=True,
        ).stdout.splitlines()

    def ref_from_rev(self, rev):
        """Determine ref from rev.

        Returns branch reference if a branch with that name exists, otherwise a
        tag reference, otherwise the rev itself (assuming it is a commit id).
        """

        if f"refs/heads/{rev}".encode() in self.refs:
            return f"refs/heads/{rev}"
        elif f"refs/remotes/origin/{rev}".encode() in self.refs:
            return f"refs/remotes/origin/{rev}"
        elif f"refs/tags/{rev}".encode() in self.refs:
            return f"refs/tags/{rev}"
        else:
            return rev

    @classmethod
    def clone_from_url(cls, url, path):
        run(
            ["git", "clone", urldefrag(url).url, path],
            cwd=Path(path).parent,
            check=True,
            capture_output=True,
        )
        return GitRepo(path)



import logging
import tarfile
import tempfile
import zipfile
from contextlib import contextmanager
from io import BytesIO
from pathlib import Path
from urllib.parse import urldefrag, urlsplit, urlunsplit

import dulwich
import requests

from .git_util import GitPath, GitRepo

logger = logging.getLogger(__name__)


def _this_or_only_subdir(path):
    members = list(path.iterdir())
    return members[0] if len(members) == 1 and members[0].is_dir() else path


@contextmanager
def _fetch_from_path(path):
    if path.is_dir():
        yield path
    else:
        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                with tarfile.open(fileobj=BytesIO(path.read_bytes())) as tar_file:
                    tar_file.extractall(path=tmp_dir)
                    yield _this_or_only_subdir(Path(tmp_dir))
            except tarfile.ReadError as error:
                logger.debug(str(error))
                try:
                    with zipfile.ZipFile(BytesIO(path.read_bytes())) as zip_file:
                        zip_file.extractall(path=tmp_dir)
                        yield _this_or_only_subdir(Path(tmp_dir))
                except zipfile.BadZipFile as error:
                    logger.debug(str(error))
                    raise RuntimeError("Failed to extract archive from file.")


@contextmanager
def _fetch_from_https(url):
    response = requests.get(url, stream=True)
    response.raise_for_status()
    content = response.content
    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_file.write(content)
        tmp_file.flush()
        try:
            with _fetch_from_path(Path(tmp_file.name)) as path:
                yield path
        except RuntimeError as error:
            raise RuntimeError(f"Unable to read from '{url}': {error}")


def _parse_git_url(git_url):
    path = urldefrag(git_url).fragment
    if "@" in git_url:
        url, rev = urldefrag(git_url).url.rsplit("@", 1)
    else:
        url, rev = urldefrag(git_url).url, None
    return url, (rev or "HEAD"), path


@contextmanager
def _fetch_from_git_https(git_url):
    url, rev, path = _parse_git_url(git_url)

    with tempfile.TemporaryDirectory() as tmp_dir:
        repo = GitRepo.clone_from_url(url, tmp_dir)
        git_path = GitPath(repo.path, repo.ref_from_rev(rev)).joinpath(path)
        with _fetch_from_path(git_path) as path:
            yield path


@contextmanager
def _fetch_from_git_local(git_url):
    url, rev, path = _parse_git_url(git_url)

    try:
        repo = GitRepo(urlsplit(url).path)
    except dulwich.errors.NotGitRepository as error:
        raise RuntimeError(
            f"{error}"
            + (" Did you use a relative path?" if urlsplit(url).netloc else "")
        )

    git_path = GitPath(repo.path, repo.ref_from_rev(rev)).joinpath(path)
    with _fetch_from_path(git_path) as path:
        yield path


@contextmanager
def fetch_from_url(url):
    ps = urlsplit(url)

    if ps.scheme in ("", "file"):  # on the local file system
        try:
            with _fetch_from_path(Path(ps.path)) as path:
                yield path
        except FileNotFoundError as error:
            raise RuntimeError(
                f"{error}" + (" Did you use a relative path?" if ps.netloc else "")
            )

    elif ps.scheme == "https":  # access archive via https
        with _fetch_from_https(url) as path:
            yield path

    elif ps.scheme == "git+https":  # clone git repository via https
        with _fetch_from_git_https(urlunsplit(ps._replace(scheme="https"))) as path:
            yield path

    elif ps.scheme == "git+file":  # parse local git repository
        with _fetch_from_git_local(urlunsplit(ps._replace(scheme=""))) as path:
            yield path

    else:  # unknown scheme
        raise ValueError(f"Unsupported scheme: {ps.scheme}.")



# -*- coding: utf-8 -*-
"""Generate the app registry website."""

import logging
import os
import os.path
import shutil
from itertools import chain
from pathlib import Path

import pkg_resources

from ..utils import parse_app_repo
from . import api, yaml
from .apps_index import generate_apps_index
from .core import AppRegistryData, AppRegistrySchemas
from .html import build_html

logger = logging.getLogger(__name__)


def copy_static_tree_from_path(base_path, static_path):

    for root, _, files in os.walk(static_path):
        # Create directory
        base_path.joinpath(Path(root).relative_to(static_path)).mkdir(
            parents=True, exist_ok=True
        )

        # Copy all files
        for filename in files:
            src = Path(root).joinpath(filename)
            dst = base_path.joinpath(Path(root).relative_to(static_path), filename)
            dst.write_bytes(src.read_bytes())
            yield dst


def _walk_pkg_resources(package, root):
    paths = pkg_resources.resource_listdir(package, root)
    for path in paths:
        dir_paths = [
            path
            for path in paths
            if pkg_resources.resource_isdir(package, os.path.join(root, path))
        ]
        yield root, list(set(paths).difference(dir_paths))
        for dir_path in dir_paths:
            yield from _walk_pkg_resources(package, os.path.join(root, dir_path))


def copy_static_tree_from_package(html_path, root="static"):
    for directory, files in _walk_pkg_resources(__package__, root):
        stem = html_path.joinpath(Path(directory).relative_to(root))
        stem.mkdir(parents=True, exist_ok=True)
        for fn in files:
            src = pkg_resources.resource_stream(
                __package__, os.path.join(directory, fn)
            )
            dst = stem.joinpath(fn)
            dst.write_bytes(src.read())
            yield dst


def build(
    apps_path: Path,
    categories_path: Path,
    base_path: Path,
    html_path: Path = None,
    api_path: Path = None,
    static_path: Path = None,
    templates_path: Path = None,
    validate_output: bool = True,
    validate_input: bool = False,
):
    """Build the app registry website (including schema files)."""
    # Parse the schemas shipped with the package.
    schemas = AppRegistrySchemas.from_package()

    # Parse the apps and categories data from the given paths.
    data = AppRegistryData(
        apps=yaml.load(apps_path),
        categories=yaml.load(categories_path),
    )
    if validate_input:
        data.validate(schemas)

    # Remove previous build (if present) and re-create root directory.
    shutil.rmtree(base_path, ignore_errors=True)
    base_path.mkdir(parents=True, exist_ok=True)

    apps_index, apps_data = generate_apps_index(
        data=data, scan_app_repository=parse_app_repo
    )

    # Build the website and API endpoints.
    logger.info(f"Building registry at: {base_path.resolve()}")
    for outfile in chain(
        # Build the html pages if the html path is specified
        chain(
            # Copy static files from package
            copy_static_tree_from_package(html_path=base_path / html_path),
            # Copy static files (if specified)
            copy_static_tree_from_path(base_path / html_path, static_path)
            if static_path
            else (),
            # Build the html pages.
            build_html(
                base_path=base_path / html_path,
                apps_index=apps_index,
                apps_data=apps_data,
                templates_path=templates_path,
            ),
        )
        if html_path is not None
        else (),
        # Build the API endpoints if the api path is specified
        api.build_api_v1(
            api_path=base_path / api_path,
            apps_index=apps_index,
            apps_data=apps_data,
            scan_app_repository=parse_app_repo,
        )
        if api_path
        else (),
    ):
        logger.info(f"  - {outfile.relative_to(base_path)}")

    if validate_output:
        if api_path:
            api.validate_api_v1(api_path=base_path / api_path, schemas=schemas)



# -*- coding: utf-8 -*-
"""Utility functions for the application registry."""

import json
import string
from pathlib import Path
from urllib.parse import urlparse


def get_html_app_fname(app_name):
    valid_characters = set(string.ascii_letters + string.digits + "_-")

    simple_string = "".join(c for c in app_name if c in valid_characters)

    return f"{simple_string}.html"


def get_hosted_on(url):
    netloc = urlparse(url).netloc

    # Remove port (if any)
    netloc = netloc.partition(":")[0]

    # Remove subdomains (this only works for domain suffixes of length 1!)
    # TODO: fix it for domains like yyy.co.uk
    netloc = ".".join(netloc.split(".")[-2:])

    return netloc


def load_json(path: Path) -> dict:
    return json.loads(path.read_text())



# -*- coding: utf-8 -*-
"""App environment specification

The specification is used to describe a reproducible environment for a
specific app, similar to the Reproducible Environment Specification (REES) [1]

[1] https://repo2docker.readthedocs.io/en/latest/specification.html

The following configuration files are recognized with the order of precedence
matching the order shown here:

.. glossary::

    setup.cfg
        Requirements listed as part of the [options]/install_requires field are
        parsed as pip-installable Python requirements for this app.

    requirements.txt
        Requirements listed within this file are parsed as pip-installable
        Python requirements for this app unless already parsed from the
        setup.cfg file.

"""
from configparser import ConfigParser
from dataclasses import dataclass, field
from typing import List

__all__ = [
    "Environment",
]


@dataclass
class Environment:
    """App environment specification.

    This dataclass contains the specification of an app environment and can be
    used to scan an existing environment configuration from a given path and to
    detect whether a given environment is meeting the specification.
    """

    python_requirements: List[str] = field(default_factory=list)

    _FILES = ("requirements.txt",)

    @staticmethod
    def _scan(path):
        def _parse_reqs(requirements):
            for line in (line.strip() for line in requirements.splitlines()):
                if line and not line.startswith("#"):
                    yield line

        def _parse_setup_cfg(setup_cfg):
            cfg = ConfigParser()
            cfg.read_string(setup_cfg)
            return cfg["options"].get("install_requires", [])

        # Parse the setup.cfg file (if present).
        try:
            yield "python_requirements", list(
                _parse_reqs(_parse_setup_cfg(path.joinpath("setup.cfg").read_text()))
            )
        except (FileNotFoundError, KeyError):
            # Parse the requirements.txt file (if present).
            try:
                yield "python_requirements", list(
                    _parse_reqs(path.joinpath("requirements.txt").read_text())
                )
            except FileNotFoundError:
                pass

    @classmethod
    def scan(cls, root):
        """Scan the root path and determine the environment specification."""
        return cls(**dict(cls._scan(root)))



# -*- coding: utf-8 -*-
"""Module that implements a basic command line interface (CLI) for AiiDAlab."""
import json
import logging
import shutil
from collections import defaultdict
from contextlib import contextmanager, nullcontext
from dataclasses import asdict
from fnmatch import fnmatch
from pathlib import Path
from textwrap import indent, wrap

import click
import pkg_resources
import requests_mock
from click_spinner import spinner
from jsonref import JsonRefError
from jsonschema.exceptions import RefResolutionError, ValidationError
from packaging.requirements import InvalidRequirement, Requirement
from packaging.version import parse
from tabulate import tabulate

from . import __version__
from .app import AppVersion
from .app import _AiidaLabApp as AiidaLabApp
from .config import AIIDALAB_APPS, AIIDALAB_REGISTRY
from .fetch import fetch_from_url
from .metadata import Metadata
from .registry import build as build_registry
from .utils import PEP508CompliantUrl, load_app_registry_index
from .utils import parse_app_repo as _parse_app_repo

SCHEMAS_CANONICAL_BASE_URL = "https://raw.githubusercontent.com/aiidalab/aiidalab/v21.10.0/aiidalab/registry/schemas"


ICON_DETACHED = "\U000025AC"  # ▬
ICON_MODIFIED = "\U00002022"  # •


@contextmanager
def _spinner_with_message(message, message_final="Done.\n", **kwargs):
    try:
        click.echo(message, err=True, nl=False)
        with spinner():
            yield
    except Exception:
        click.echo()  # create new line
        raise
    else:
        click.echo(message_final, err=True)


def _list_apps(apps_path):
    if apps_path.is_dir():
        for path in apps_path.iterdir():
            if path.is_dir():
                if path.stem.startswith(".") or path.stem == "__pycache__":
                    continue  # exclude hidden directories and the Python cache directory.
                app_name = str(path.relative_to(apps_path))
                try:
                    yield path, app_name, AiidaLabApp.from_id(app_name)
                except KeyError:
                    yield path, app_name, None
    elif apps_path.exists():
        raise click.ClickException(
            f"The apps path ('{apps_path}') appears to not be a valid directory."
        )


@click.group()
@click.version_option(version=__version__, prog_name="AiiDAlab")
@click.option("-v", "--verbose", count=True)
def cli(verbose):
    logging.basicConfig(level=max(0, logging.WARNING - 10 * verbose))


@cli.command()
def info():
    """Show information about the AiiDAlab configuration."""
    click.echo(f"AiiDAlab, version {__version__}")
    click.echo(f"Apps path:      {Path(AIIDALAB_APPS).resolve()}")
    click.echo(f"Apps registry:  {AIIDALAB_REGISTRY}")


@cli.command()
@click.argument("app-query", default="*")
@click.option(
    "--pre",
    "prereleases",
    is_flag=True,
    # The default value for this flag is 'None', because we do not explicitly
    # _disallow_ prereleases when the option is not set, but it depends on the
    # provided requirement. That means when the query includes a prerelease
    # version (e.g. my-app>=1.0rc), then results will include prereleases,
    # whether this option is set or not. Otherwise, the results will only
    # include prereleases when this option is set.
    default=None,
    help="Include prereleases among the search candidates. Note: Prereleases "
    "are automatically included if an app has only prereleases.",
)
def search(app_query, prereleases):
    """Search for apps within the registry.

    Accepts either a search query with app names that contain wildcards ('*', '?')
    or a specific app requirement (e.g. 'hello-world>1.0').

    Examples:

    Find all apps where the name starts with "hello":

        search 'hello*'

    Find all apps with name "hello-world" with version greater or equal than 1.0:

        search 'hello-world>=1.0'

    """

    with _spinner_with_message("Collecting apps and releases... "):
        try:
            app_requirements = [Requirement(app_query)]
        except InvalidRequirement:  # interpreted as general search query
            try:
                registry = load_app_registry_index()
            except RuntimeError as error:
                raise click.ClickException(error)
            app_requirements = [
                Requirement(app_name)
                for app_name in registry["apps"].keys()
                if fnmatch(app_name, app_query)
            ]

    for app_requirement in app_requirements:
        try:
            app = AiidaLabApp.from_id(app_requirement.name)
        except KeyError:
            raise click.ClickException(
                f"Did not find entry for app with name '{app_requirement.name}'."
            )
        matching_releases = [
            version
            for version in app.find_matching_releases(
                app_requirement.specifier, prereleases
            )
        ]
        if matching_releases:
            click.echo(
                "\n".join(f"{app.name}=={version}" for version in matching_releases)
            )


def _tabulate_apps(apps, headers=("App name", "Version", "Path"), **kwargs):
    rows = []
    for app_path, app_name, app in sorted(apps):
        if app is None:
            version = ICON_DETACHED
        else:
            version = f"{app.installed_version()}{ICON_MODIFIED if app.dirty() else ''}"
        rows.append([app_name, version, str(app_path)])
    yield tabulate(rows, headers=headers, colalign=("left", "left", "left"), **kwargs)
    if rows:
        yield f"\n{ICON_DETACHED}:detached {ICON_MODIFIED}:modified"


@cli.command(name="list")
def list_apps():
    """List all installed apps.

    This command will list all apps, their version, and their full path.
    """

    apps = list(_list_apps(Path(AIIDALAB_APPS)))
    if len(apps) > 0:
        click.echo("\n".join(_tabulate_apps(apps)))
    else:
        click.echo("No apps installed.", err=True)


def _parse_requirement(app_requirement):
    try:
        return Requirement(app_requirement)
    except InvalidRequirement as error:
        from urllib.parse import urlsplit

        if urlsplit(app_requirement).scheme:
            raise click.ClickException(
                "It looks like you tried to install directly from a PEP 508 "
                "compliant URL. Make sure to use the 'app-name@url' syntax "
                "to provide a name for the app."
            )
        raise click.ClickException(
            f"Invalid requirement '{app_requirement}': {error!s}\n"
        )


def _find_registered_app_from_id(name):
    """Find app for a given requirement."""
    try:
        app = AiidaLabApp.from_id(name)
        if app.is_registered:
            return app
        else:
            raise click.ClickException(
                f"App '{app}' was installed locally and/or is not registered."
            )
    except KeyError:
        raise click.ClickException(f"Did not find entry for app with name '{name}'.")


def _find_app_and_releases(app_requirement):
    """Find app and a suitable release for a given requirement."""
    app = _find_registered_app_from_id(app_requirement.name)
    matching_releases = app.find_matching_releases(app_requirement.specifier)
    return app, matching_releases


@cli.command()
@click.argument("app-requirement", nargs=-1)
@click.option(
    "--indent", type=int, help="Specify level of identation of the JSON output."
)
def show_environment(app_requirement, indent):
    """Show environment specification of apps.

    Example:

    Show the environment specification for the latest version of the hello-world app:

        show-environment hello-world

    Show the combined environment for version 0.1.0 of the 'hello-world'
    app and 0.12-compatible release of 'other-app':

        show-environment hello-world==0.1.0 other-app~=0.12

    """

    with _spinner_with_message(
        "Collecting apps and their environment specifications... "
    ):
        apps_and_releases = {
            requirement: _find_app_and_releases(requirement)
            for requirement in map(_parse_requirement, app_requirement)
        }

    if not len(apps_and_releases):
        # We show a warning if no apps are selected, but we still show the JSON
        # environment specification. Less potential to break scripted pipelines
        # that might operate on zero or more selected apps.
        click.secho("No apps selected.", err=True, fg="yellow")

    aggregated_environment = defaultdict(list)
    for requirement, (app, selected_versions) in apps_and_releases.items():
        try:
            environment = app.releases[selected_versions[0]].get("environment")
        except IndexError:  # no matching release
            raise click.ClickException(
                f"{app.name}: No matching release for '{requirement.specifier}'. "
                f"Available releases: {','.join(map(str, sorted(map(parse, app.releases))))}"
            )
        else:
            click.echo(f"Selected '{app.name}=={selected_versions[0]}'", err=True)
            if environment:
                for key in environment:
                    aggregated_environment[key].extend(environment[key])
            else:
                click.secho(
                    f"No environment declared for '{app.name}'.",
                    fg="yellow",
                )

    click.echo(json.dumps(aggregated_environment, indent=indent))


def _find_version_to_install(
    app_requirement, force, dependencies, python_bin, prereleases
):
    if app_requirement.url is not None:
        with fetch_from_url(app_requirement.url) as repo:
            metadata = Metadata.parse(repo)
            registry_entry = dict(name=app_requirement.name, metadata=asdict(metadata))
            app = AiidaLabApp.from_id(
                app_requirement.name, registry_entry=registry_entry
            )
            if not (force or (dependencies in ("install", "ignore"))):
                raise click.ClickException(
                    f"Unable to check compatibility for {app_requirement} prior to installation."
                )

        if force or not app.is_installed():
            return app, PEP508CompliantUrl(app_requirement.url)
        else:
            return app, None

    app = _find_registered_app_from_id(app_requirement.name)

    assert dependencies in ("install", "require", "ignore")
    matching_releases = app.find_matching_releases(
        app_requirement.specifier, prereleases
    )
    compatible_releases = [
        version
        for version in matching_releases
        if force
        or (dependencies in ("install", "ignore"))
        or app.is_compatible(version, python_bin=python_bin)
    ]

    # Best case scenario: matching and compatible release found!
    if compatible_releases:
        version_to_install = compatible_releases[0]
        if force or version_to_install != app.installed_version():
            return app, version_to_install
        else:
            return app, None  # app already installed in that version

    # There are matching releases, but none are compatible, inform user.
    elif len(matching_releases) > 0:
        raise click.ClickException(
            f"There are releases matching '{app_requirement}' ("
            f"{','.join(map(str, sorted(map(parse, matching_releases))))}), however "
            "none of these are compatible with this environment."
        )

    # No matching releases, inform user about available releases.
    else:
        raise click.ClickException(
            f"No matching release for '{app_requirement}'. "
            f"Available releases: {','.join(map(str, sorted(map(parse, app.releases))))}"
        )


@cli.command()
@click.argument("app-requirement", nargs=-1)
@click.option("--yes", is_flag=True, help="Do not prompt for confirmation.")
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Show what app (version) would be installed, but do not actually perform the installation.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Ignore all warnings and perform potentially dangerous operations anyways.",
)
@click.option(
    "-d",
    "--dependencies",
    type=click.Choice(["install", "require", "ignore"]),
    default="install",
    help="Select whether to install app dependencies (the default), "
    "require them to be installed prior to installation, "
    "or completely ignore them.",
    show_default=True,
)
@click.option(
    "--python",
    "python_bin",
    default=shutil.which("python"),
    type=click.Path(dir_okay=False, exists=True),
    help="The Python binary to use to install Python dependencies or to check their presence.",
    show_default=True,
)
@click.option(
    "--pre",
    "prereleases",
    is_flag=True,
    # The default value for this flag is 'None', because we do not explicitly
    # _disallow_ prereleases when the option is not set, but it depends on the
    # provided requirement. That means when the query includes a prerelease
    # version (e.g. my-app>=1.0rc), then the list of potential installation
    # candidates will include prereleases regardless of whether this option is
    # set or not.  Otherwise, prereleases are only considered when this option
    # is set.
    default=None,
    help="Include prereleases among the candidates for installation.",
)
def install(
    app_requirement, yes, dry_run, force, dependencies, python_bin, prereleases
):
    """Install apps.

    This command will install the latest version of an app matching the given requirement.

    Examples:

    Install the 'hello-world' app with the latest version that matches the specification '>=1.0':

        install hello-world>=1.0

    Install the 'hello-world' app directly with a PEP 508 compliant URL:

        install hello-world@git+https://github.com/aiidalab/aiidalab-hello-world.git

    Note: It is necessary to explicitly specify the app name before the '@'.
    """
    with _spinner_with_message("Collecting apps matching requirements... "):
        install_candidates = {
            requirement: _find_version_to_install(
                requirement,
                dependencies=dependencies,
                force=force,
                python_bin=python_bin,
                prereleases=prereleases,
            )
            for requirement in map(_parse_requirement, set(app_requirement))
        }

    if all(version is None for (_, version) in install_candidates.values()):
        click.echo("Nothing to install, exiting.", err=True)
        return

    elif any(version is None for (_, version) in install_candidates.values()):
        for requirement, (_, version) in install_candidates.items():
            if version is None:
                click.secho(
                    f"Requirement '{requirement}' already satisfied.",
                    err=True,
                    fg="yellow",
                )
        click.secho(
            "Use the '-f/--force' option to re-install anyways.\n",
            err=True,
            fg="yellow",
        )

    apps_table = tabulate(
        [
            [app.name, version, app.path]
            for _, (app, version) in install_candidates.items()
            if version is not None
        ],
        headers=["App", "Version", "Path"],
    )
    click.echo(f"Would install:\n\n{indent(apps_table, '  ')}\n")
    if yes or click.confirm("Proceed?", default=True):
        for app, version in install_candidates.values():
            if version is not None:
                if dry_run:
                    click.secho(
                        f"Would install '{app.name}' version '{version}'.", fg="green"
                    )
                else:
                    try:
                        app.install(
                            version=version,
                            install_dependencies=dependencies == "install",
                            python_bin=python_bin,
                        )
                    except RuntimeError as error:
                        click.secho(f"Error: {error}", fg="red")
                        break
                    else:
                        click.secho(
                            f"Installed '{app.name}' version '{version}'.", fg="green"
                        )


@cli.command()
@click.argument("app-name", nargs=-1)
@click.option("--yes", is_flag=True, help="Do not prompt for confirmation.")
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Show what apps would be uninstalled, but do not actually perform the operation.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Ignore all warnings and perform potentially dangerous operations anyways.",
)
def uninstall(app_name, yes, dry_run, force):
    """Uninstall apps."""
    with _spinner_with_message("Collecting apps to uninstall... "):
        apps_to_uninstall = [
            (path, name, app)
            for (path, name, app) in _list_apps(Path(AIIDALAB_APPS))
            if any(fnmatch(name, app_name_) for app_name_ in app_name)
        ]

        # Determine whether some apps are dirty or detached.
        dirty = {
            name: (app is not None and app.dirty())
            for _, name, app in apps_to_uninstall
        }

        detached = {
            name: app is None or app.installed_version() is AppVersion.UNKNOWN
            for _, name, app in apps_to_uninstall
        }

    if not len(apps_to_uninstall):
        click.echo("Nothing to uninstall, exiting.", err=True)
        return

    apps_table = "\n".join(_tabulate_apps(apps_to_uninstall, tablefmt="simple"))
    click.echo(f"Would uninstall:\n\n{indent(apps_table, '  ')}\n")

    if any(dirty.values()):
        click.secho(
            "\n".join(
                wrap(
                    "WARNING: Some applications selected for removal have modifications "
                    "that would potentially be lost."
                )
            ),
            err=True,
            fg="yellow" if force else "red",
        )

    if any(detached.values()):
        click.secho(
            "\n".join(
                wrap(
                    "WARNING: Some applications selected for removal are detached, meaning they can "
                    "either not be found in the registry or are installed with unknown versions. "
                    "Their removal may lead to data loss."
                )
            ),
            err=True,
            fg="yellow" if force else "red",
        )

    potential_data_loss = any(dirty.values()) or any(detached.values())
    if potential_data_loss and not force:
        click.secho(
            "Use the -f/--force option to ignore all warnings.", err=True, fg="red"
        )
        raise click.Abort()

    if yes or click.confirm("Proceed?", default=not potential_data_loss):
        for app_path, _, app in apps_to_uninstall:
            if app is None:
                if dry_run:
                    click.secho(f"Would remove directory '{app_path!s}'.", err=True)
                else:
                    shutil.rmtree(app_path)
                    click.echo(f"Removed directory '{app_path!s}'.", err=True)
            else:
                if dry_run:
                    click.echo(
                        f"Would uninstall '{app.name}' ('{app.path!s}').", err=True
                    )
                else:
                    app.uninstall()
                    click.echo(
                        f"Uninstalled '{app.name}' ('{app.path!s}').",
                        err=True,
                    )


@contextmanager
def _mock_schemas_endpoints():
    schema_paths = [
        path
        for path in pkg_resources.resource_listdir(f"{__package__}.registry", "schemas")
        if path.endswith(".schema.json")
    ]

    with requests_mock.Mocker(real_http=True) as mocker:
        for schema_path in schema_paths:
            schema = json.loads(
                pkg_resources.resource_string(
                    f"{__package__}.registry", f"schemas/{schema_path}"
                )
            )
            mocker.get(
                schema.get("$id", f"{SCHEMAS_CANONICAL_BASE_URL}/{schema_path}"),
                text=json.dumps(schema),
            )
        yield


@cli.group()
def registry():
    """Functions related to managing an app registry."""


@registry.command()
@click.argument("url")
def parse_app_repo(url):
    """Parse an app repo for metadata and other information.

    Use this command to parse a local or remote app repository for the app
    metadata and environment specification.

    Examples:

    For a local app repository, provide the absolute or relative path:

        parse-app-repo /path/to/aiidalab-hello-world

    For a remote app repository, provide a PEP 508 compliant URL, for example:

        parse-app-repo git+https://github.com/aiidalab/aiidalab-hello-world.git@v1.0.0
    """
    click.echo(f"Parsing {url} ...", err=True)
    try:
        click.echo(json.dumps(_parse_app_repo(url)))
    except (ValueError, TypeError) as error:
        click.secho(
            f"Failed to parse metadata from '{url}': {error!s}",
            err=True,
            fg="red",
        )


@registry.command()
@click.option(
    "--apps",
    type=click.Path(exists=True, dir_okay=False),
    default="apps.yaml",
    help="Path to the registry's apps.yaml file.",
    show_default=True,
)
@click.option(
    "--categories",
    type=click.Path(exists=True, dir_okay=False),
    default="categories.yaml",
    help="Path to the registry's categories.yaml file.",
    show_default=True,
)
@click.option(
    "-o",
    "--out",
    type=click.Path(file_okay=False, writable=True),
    default="build/",
    show_default=True,
    help="Path to the base directory of where the registry's HTML and API pages are generated.",
)
@click.option(
    "--html-path",
    type=str,
    help="Relative path to the build directory at which the HTML pages are generated. "
    "Set to an empty value to skip the build of HTML pages entirely.",
    default=".",
    show_default=True,
)
@click.option(
    "--api-path",
    type=str,
    help="Relative path to the build directory at which the API pages are generated. "
    "Set to an empty value to skip the build of API pages entirely.",
    default="api/v1",
    show_default=True,
)
@click.option("--static", type=click.Path(file_okay=False, exists=True))
@click.option("-t", "--templates", type=click.Path(file_okay=False, exists=True))
@click.option(
    "--validate/--no-validate",
    is_flag=True,
    default=True,
    show_default=True,
    help="Validate inputs and outputs against the published or local schemas.",
)
@click.option(
    "-m",
    "--mock-schemas-endpoints",
    "mock_schemas",
    is_flag=True,
    help="Mock the schemas endpoints such that the local versions are used insted of the published ones.",
)
def build(
    apps,
    categories,
    out,
    html_path,
    api_path,
    static,
    templates,
    validate,
    mock_schemas,
):
    """Build the app store website and API endpoints.

    Example:

        build --apps=apps.yaml --categories=categories.yaml --out=./build/
    """
    if any(path and Path(path).is_absolute() for path in (html_path, api_path)):
        raise click.ClickException("The html- and api-paths must be relative paths.")

    maybe_mock = _mock_schemas_endpoints() if mock_schemas else nullcontext()
    with maybe_mock:
        try:
            build_registry(
                apps_path=Path(apps),
                categories_path=Path(categories),
                base_path=Path(out),
                html_path=Path(html_path) if html_path else None,
                api_path=Path(api_path) if api_path else None,
                static_path=Path(static) if static else None,
                templates_path=Path(templates) if templates else None,
                validate_output=validate,
                validate_input=validate,
            )
        except ValidationError as error:
            raise click.ClickException(f"Error during schema validation:\n{error}")
        except RefResolutionError as error:
            raise click.ClickException(
                f"Failed to resolve schema JSON reference: {error}\n\n"
                "Maybe try to run with `--mock-schemas-endpoints` ?"
            )
        except JsonRefError as error:
            raise click.ClickException(
                f"Failed to resolve JSON reference: {error.reference}\n{error.cause}"
            )


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from pathlib import Path
from urllib.parse import urlsplit

import cachecontrol
import jsonref
import requests
from ruamel.yaml import YAML

REQUESTS = cachecontrol.CacheControl(requests.Session())


class JsonYamlLoader(jsonref.JsonLoader):

    safe_yaml = YAML(typ="safe")

    def __call__(self, uri, **kwargs):
        uri_split = urlsplit(uri)
        if Path(uri_split.path).suffix in (".yml", ".yaml"):
            if uri_split.scheme == "file":
                content = Path(uri_split.path).read_bytes()
            else:
                response = REQUESTS.get(uri)
                response.raise_for_status()
                content = response.content
            return self.safe_yaml.load(content)
        else:
            return super().__call__(uri, **kwargs)


json_yaml_loader = JsonYamlLoader()


def replace_refs(obj):
    """Dereference all references in obj.

    References may point to JSON or YAML files.
    """
    return jsonref.JsonRef.replace_refs(obj, loader=json_yaml_loader)


def loads(s):
    """Deserialize serialized YAML file 's' to a Python object and dereference all references."""
    return replace_refs(YAML(typ="safe").load(s))


def load(path):
    """Deserialize YAML file at path to a Python object and dereference all references."""
    return loads(Path(path).read_text())


__all__ = [
    "load",
    "loads",
    "replace_refs",
]



# -*- coding: utf-8 -*-
"""Generate the app registry website HTML pages."""
from collections import defaultdict

from jinja2 import (
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
    select_autoescape,
)

from . import template_filters


def build_html(base_path, apps_index, apps_data, templates_path):
    """Generate the app registry website at the base_path path."""

    # Create base_path directory if needed
    base_path.mkdir(parents=True, exist_ok=True)

    # Setup template environment
    loaders = [PackageLoader(__name__)]
    if templates_path:
        loaders.insert(0, FileSystemLoader(templates_path))

    env = Environment(
        loader=ChoiceLoader(loaders),
        autoescape=select_autoescape(["html", "xml"]),
    )
    env.filters["sort_semantic"] = template_filters.sort_semantic

    app_page_template = env.get_template("app_page.html")
    main_index_template = env.get_template("index.html")

    # Make single-entry pages based on app_page.html
    base_path.joinpath("apps").mkdir()
    html_template_data = defaultdict(dict)

    for app_id in apps_index["apps"]:
        subpage = base_path.joinpath("apps", app_id, "index.html")
        html_template_data[app_id]["subpage"] = str(subpage.relative_to(base_path))
        html_template_data[app_id]["metadata"] = apps_data[app_id]["metadata"]
        html_template_data[app_id]["releases"] = apps_data[app_id]["releases"]

        subpage.parent.mkdir()
        subpage.write_text(
            app_page_template.render(
                category_map=apps_index["categories"], **html_template_data[app_id]
            ),
            encoding="utf-8",
        )
        yield subpage

    # Make index page based on main_index.html
    rendered = main_index_template.render(apps=html_template_data)
    outfile = base_path / "index.html"
    outfile.write_text(rendered, encoding="utf-8")
    yield outfile



# -*- coding: utf-8 -*-
"""Manage a registry of applications."""

from .core import AppRegistryData, AppRegistrySchemas
from .web import build, build_html

__all__ = [
    "AppRegistryData",
    "AppRegistrySchemas",
    "build",
    "build_html",
]



# -*- coding: utf-8 -*-
import logging
import os
import re
from dataclasses import dataclass, replace
from urllib.parse import urlsplit, urlunsplit

from aiidalab.environment import Environment
from aiidalab.fetch import fetch_from_url
from aiidalab.git_util import GitRepo
from aiidalab.metadata import Metadata

logger = logging.getLogger(__name__)


@dataclass
class Release:
    environment: Environment
    metadata: Metadata
    url: str


RELEASE_LINE_PATTERN = r"^(?P<rev>[^:]*?)(:(?P<rev_selection>.*))?$"


def _split_release_line(url):
    parsed_url = urlsplit(url)
    if "@" in parsed_url.path:
        path, release_line = parsed_url.path.rsplit("@", 1)
        return urlunsplit(parsed_url._replace(path=path)), release_line
    return url, None


def _get_release_commits(repo, release_line):
    match = re.match(RELEASE_LINE_PATTERN, release_line)

    if not match:
        raise ValueError(f"Invalid release line specification: {release_line}")

    rev = match.groupdict()["rev"] or repo.get_current_branch()

    if match.groupdict()["rev_selection"] is None:
        # No rev_selection means to select this and only this specific
        # revision.  For example: '@main' means, simply checkout 'main' (could
        # be a branch or a tag, however branches have priority).
        for ref in [
            f"refs/heads/{rev}",
            f"refs/remotes/origin/{rev}",
            f"refs/tags/{rev}",
        ]:
            if ref.encode() in repo.refs:
                yield rev, repo.get_peeled(ref.encode()).decode()
                return
        # rev likely committish (commit)
        yield rev, rev

    elif match.groupdict()["rev_selection"]:
        # A rev selection is provided, we fetch the full rev list for the given
        # selection.  For example: '@main:v1..v2' means all commits from v1
        # (exclusive) to v2 (inclusive).

        rev_selection = match.groupdict()["rev_selection"]

        # While the git rev-list command supports listing revisions for a
        # single ref, in this context we only support rev selections for a
        # range, not for individual refs.
        if ".." not in match.groupdict()["rev_selection"]:
            raise ValueError(
                "The rev_selection '{rev_selection}' must specify a range, "
                "that means must contain the range operator '..'."
            )

        # Incomplete revision selections such as `@main:v1..` must be expanded to
        # `@main:v1..main`. Therefore, we first determine the branch ref for the
        # given rev:
        for ref in [f"refs/heads/{rev}", f"refs/remotes/origin/{rev}"]:
            if ref.encode() in repo.refs:
                break
        else:
            raise RuntimeError(f"Revision '{rev}' not a valid branch name.")

        # Transform a potentially incomplete rev_selection into one that
        # contains the branch ref. For example `v1..` is expanded to
        # `v1..{ref}`, where `{ref}` is replaced with the actual reference.
        start, _, stop = rev_selection.rpartition("..")
        selected_commits = repo.rev_list(f"{start or ref}..{stop or ref}")

        for tag in repo.get_merged_tags(rev):
            commit = repo.get_commit_for_tag(tag)
            if commit in selected_commits:
                yield tag, commit

    else:
        # The rev selection is empty, select all tagged commits for the
        # selected revision.  For example: '@main:' means all tagged commits on
        # the main branch.
        for tag in repo.get_merged_tags(rev):
            yield tag, repo.get_commit_for_tag(tag)


def _gather_releases(release_specs, scan_app_repository, app_metadata):
    for release_spec in release_specs:
        if isinstance(release_spec, str):
            url = release_spec
            environment_override = None
            metadata_override = None
            version_override = None
        else:
            url = release_spec["url"]
            environment_override = release_spec.get("environment")
            metadata_override = release_spec.get("metadata", app_metadata)
            version_override = release_spec.get("version")

        def _set_overrides(version, release):
            return version_override or version, replace(
                release,
                environment=environment_override or release.environment,
                metadata=metadata_override or release.metadata,
            )

        # The way that an app is retrieved is determined by the scheme of the
        # release url.  For example, "git+https://example.com/my-app.git" means
        # that the app is located at a remote git repository from which it can
        # be downloaded (cloned) via https.
        base_url, release_line = _split_release_line(url)
        parsed_url = urlsplit(base_url)

        with fetch_from_url(base_url) as repo_path:
            if parsed_url.scheme.startswith("git+"):
                repo = GitRepo(os.fspath(repo_path))
                for ref, sha in _get_release_commits(
                    repo, release_line or repo.get_current_branch()
                ):
                    # Parse environment from local copy of repository.
                    metadata_and_environment = scan_app_repository(
                        f"git+file:{os.fspath(repo_path.resolve())}@{sha}"
                    )
                    if (
                        metadata_and_environment["metadata"] is None
                        and metadata_override is None
                    ):
                        logger.warning(
                            f"Failed to parse metadata for {base_url}@{ref} and no override specified, skipping release!"
                        )
                        continue

                    # Replace release specifier to point to specific commit.
                    path = f"{parsed_url.path.rsplit('@', 1)[0]}@{sha}"
                    release = Release(
                        url=urlunsplit(parsed_url._replace(path=path)),
                        **metadata_and_environment,
                    )
                    yield _set_overrides(ref, release)
            else:
                release = Release(
                    url=url,
                    **scan_app_repository(f"file:{os.fspath(repo_path.resolve())}"),
                )
                yield _set_overrides(None, release)


def gather_releases(app_data, scan_app_repository):
    for version, release in _gather_releases(
        app_data.get("releases", []),
        scan_app_repository,
        app_metadata=app_data.get("metadata"),
    ):
        if version is None:
            raise ValueError(f"Unable to determine version for: {release}")
        yield version, release



# -*- coding: utf-8 -*-
"""Generate the apps index including all aggregated metadata."""
import logging
from collections import OrderedDict
from copy import deepcopy
from dataclasses import asdict

import jsonschema
from packaging.version import parse

from . import util
from .releases import gather_releases

logger = logging.getLogger(__name__)


def _determine_app_name(app_id):
    """Currently the app name is identical to its id."""
    assert util.get_html_app_fname(app_id) == f"{app_id}.html"
    return app_id


def _migrate_app_data(app_data):
    if "metadata" in app_data:
        # Set defaults
        app_data["metadata"].setdefault("categories", app_data.pop("categories", []))
        app_data["metadata"].setdefault("logo", app_data.pop("logo", None))

        # Remove deprecated keys from app metadata.
        for key in (
            "requires",
            "version",
        ):
            if key in app_data["metadata"]:
                del app_data["metadata"][key]


def _fetch_app_data(app_id, app_data, scan_app_repository):
    # Gather all release data.
    _migrate_app_data(app_data)

    app_data["name"] = _determine_app_name(app_id)
    app_data["releases"] = {
        version: asdict(release)
        for version, release in gather_releases(app_data, scan_app_repository)
    }
    if len(app_data["releases"]):
        # Sort all releases semantically to determine the latest version.
        latest_version = sorted(app_data["releases"], key=parse, reverse=True)[0]
        # The metadata of the latest release is considered authoritative for the whole app.
        app_data["metadata"] = app_data["releases"][latest_version]["metadata"]
        return app_data


def generate_apps_index(data, scan_app_repository):
    """Generate the comprehensive app index.

    This index is built from the apps data and includes additional information
    such as the releases.
    """

    apps_data = OrderedDict()
    index = {
        "apps": OrderedDict(),
        "categories": data.categories,
    }
    logger.info("Fetching app data...")

    for app_id in sorted(data.apps.keys()):
        logger.info(f"  - {app_id}")
        app_data = _fetch_app_data(
            app_id, deepcopy(data.apps[app_id]), scan_app_repository
        )
        if app_data is not None:  # would be None if the app had no release yet.
            apps_data[app_id] = app_data
            index["apps"][app_id] = {
                "name": app_data["name"],
                "categories": app_data["metadata"].get("categories", []),
            }

    return index, apps_data


def validate_apps_index_and_apps(apps_index, apps_index_schema, apps, app_schema):
    """Validate the apps_index file."""

    # Validate apps index against schema
    jsonschema.validate(instance=apps_index, schema=apps_index_schema)

    # Validate index categories
    for apps_index_entry in apps_index["apps"].values():
        for category in apps_index_entry["categories"]:
            assert category in apps_index["categories"]

    # Validate all apps
    for app in apps:
        jsonschema.validate(instance=app, schema=app_schema)



# -*- coding: utf-8 -*-
"""App metadata specification"""
from configparser import ConfigParser
from dataclasses import dataclass, field
from typing import List

__all__ = [
    "Metadata",
]


def _map_development_state(classifiers):
    "Map standard trove classifiers (PEP 301) to aiidalab development states."
    if "Development Status :: 1 - Planning" in classifiers:
        return "registered"
    elif "Development Status :: 5 - Production/Stable" in classifiers:
        return "stable"
    elif any(
        classifier in classifiers
        for classifier in (
            "Development Status :: 2 - Pre-Alpha",
            "Development Status :: 3 - Alpha",
            "Development Status :: 4 - Beta",
        )
    ):
        return "development"
    else:
        return "registered"


def _parse_config_dict(dict_):
    "Parse a config dict string for key-values pairs."
    for line in dict_.splitlines():
        if line:
            key, value = line.split("=")
            yield key.strip(), value.strip()


def _parse_setup_cfg(setup_cfg):
    "Parse a setup.cfg configuration file string for metadata."
    cfg = ConfigParser()
    cfg.read_string(setup_cfg)

    try:
        metadata_pep426 = cfg["metadata"]
    except KeyError:
        metadata_pep426 = dict()
    aiidalab = cfg["aiidalab"] if "aiidalab" in cfg else dict()

    yield "title", aiidalab.get("title", metadata_pep426.get("name"))
    yield "version", aiidalab.get("version", metadata_pep426.get("version"))
    yield "description", aiidalab.get("description", metadata_pep426.get("description"))
    yield "authors", aiidalab.get("authors", metadata_pep426.get("author"))
    yield "external_url", aiidalab.get("external_url", metadata_pep426.get("url"))

    project_urls = dict(_parse_config_dict(metadata_pep426.get("project_urls", "")))
    yield "documentation_url", aiidalab.get(
        "documentation_url",
        project_urls.get("Documentation") or project_urls.get("documentation"),
    )
    yield "logo", aiidalab.get(
        "logo", project_urls.get("Logo") or project_urls.get("logo")
    )
    yield "state", aiidalab.get(
        "state", _map_development_state(metadata_pep426.get("classifiers", []))
    )

    yield "categories", aiidalab.get("categories", [])


@dataclass
class Metadata:
    """App metadata specification."""

    title: str
    description: str
    authors: str = None
    state: str = None
    documentation_url: str = None
    external_url: str = None
    logo: str = None
    categories: List[str] = field(default_factory=list)
    version: str = None

    _search_dirs = (".aiidalab", "./")

    @staticmethod
    def _parse(path):

        try:
            return {
                key: value
                for key, value in _parse_setup_cfg(
                    path.joinpath("setup.cfg").read_text()
                )
                if value is not None
            }
        except FileNotFoundError:
            return {}

    @classmethod
    def parse(cls, root):
        """Parse the app metadata from a setup.cfg within the app repository.

        This function will parse metadata fields from a possible "aiidalab"
        section, but falls back to the standard fields defined as part of the
        PEP 426-compliant metadata section for any missing values.
        """
        for path in (root.joinpath(dir_) for dir_ in cls._search_dirs):
            if path.is_dir():
                return cls(**dict(cls._parse(path)))

        raise ValueError(f"Directory '{root}' does not exist.")



# -*- coding: utf-8 -*-
"""Custom filters for the Jinja2 template engine."""
from packaging.version import parse


def sort_semantic(versions, pre=False):
    return [
        version
        for version in sorted(versions, key=parse, reverse=True)
        if pre or not parse(version).is_prerelease
    ]



# -*- coding: utf-8 -*-
"""Generate API endpoints."""
import json

from .apps_index import validate_apps_index_and_apps


def build_api_v1(api_path, apps_index, apps_data, scan_app_repository):
    """Build tree for API endpoint v1."""

    # Create base path if necessary.
    api_path.mkdir(parents=True, exist_ok=True)

    # Write apps_index.json file.
    outfile = api_path / "apps_index.json"
    rendered = json.dumps(apps_index, ensure_ascii=False)
    outfile.write_text(rendered, encoding="utf-8")
    yield outfile

    api_path.joinpath("apps").mkdir()
    for app_id, app_data in apps_data.items():
        # Write apps/{appId}.json
        outfile = api_path / "apps" / f"{app_id}.json"
        rendered = json.dumps(app_data, ensure_ascii=False)
        outfile.write_text(rendered, encoding="utf-8")
        yield outfile


def validate_api_v1(api_path, schemas):
    """Validate tree for API endpoint v1."""
    apps_index = json.loads(api_path.joinpath("apps_index.json").read_text())
    apps = [
        json.loads(api_path.joinpath("apps", f"{app_id}.json").read_text())
        for app_id in apps_index["apps"]
    ]
    validate_apps_index_and_apps(
        apps_index,
        apps_index_schema=schemas.apps_index,
        apps=apps,
        app_schema=schemas.app,
    )
